import re
import logging
import contextlib

from typing import Optional, Union

import pyodbc

from . import exceptions
from ._regex_lookup_dict import RegexLookupDict

logger = logging.getLogger(__name__)

EXECUTE_ERRORS = RegexLookupDict(
    {
        "invalid object name": exceptions.InvalidSQLQuery(
            "An invalid object name was used in the provided SQL Query. "
            "Please check the table names and schema supplied."
        ),
        "invalid column name": exceptions.InvalidSQLQuery("An invalid column name was used in the provided SQL Query."),
        "the select permission was denied": exceptions.PermissionDenied(
            "The user provided for the server connection does not have read access"
            "to all the tables specified in the SQL Query."
        ),
        "not able to access the database": exceptions.PermissionDenied(
            "The user provided for the server connection does not have read access to the database specified."
        ),
        r"the sql contains \d+ parameter markers, but \d+ parameters were supplied": exceptions.InvalidSQLQuery(
            "Not enough values were supplied for all placeholders in the SQL Query."
        ),
    }
)


@contextlib.contextmanager
def _connect(
    *,
    address: str,
    default_database: str,
    username: str,
    password: str,
    instance: Optional[str],
    port: Optional[Union[str, int]],
    freetds_driver="/usr/lib/libtdsodbc.so.0",
):
    connection = None

    server = f"{address}\\{instance}" if instance is not None else address
    connection_kwargs = {
        "driver": f"{{{freetds_driver}}}",
        "server": server,
        "database": default_database,
        "uid": username,
        "password": password,
        "tds_version": "8.0",
    }
    # Only add the port if it is supplied - including a None default value will break things.
    if port is not None:
        connection_kwargs.update({"port": port})  # type: ignore

    connection_display = (
        f"(server: {address}, database: {default_database}, instance: {instance or 'default'}, "
        f"port: {port if port is not None else 'default'})"
    )
    logger.info(f"opening connection: {connection_display}")
    try:
        connection = pyodbc.connect(**connection_kwargs)
        yield connection
    except (pyodbc.ProgrammingError, pyodbc.Error) as e:
        if all((port is not None, instance is not None)):
            logger.error(
                "Both port and instance name were supplied. If you are trying to connect to a "
                "non-default instance on this server you may only need to supply the instance name."
            )
        error_message = str(e).lower()

        if "login failed for user" in error_message:
            # This can happen when:
            #
            # - Local SQL legitimately fails
            # - The database does not exist
            # - The instance & port combination is incorrect
            #
            # It never happens when AD authentication legitimiately fails (great)
            logger.error(str(e))
            if "\\" in username:
                # Domain account, correct credentials, most likely bad database name or instance & port combo...
                exception = exceptions.ConnectionFailed(  # type: ignore
                    f"Failed to connect to server. This may be because the database does not exist, "
                    f"or the combination of instance & port is incorrect {connection_display}."
                )
            else:
                # Local account, could be credentials or database name or the instance & port combo. Without
                # knowing the default instance or default database, we can't be any clearer than this...
                exception = exceptions.AuthenticationFailed(  # type: ignore
                    f"Failed to authenticate user {username}. This may be because the credentials, database name, "
                    f"or combination of instance & port is incorrect {connection_display}."
                )
            raise exception from None

        elif "login is from an untrusted domain" in error_message:
            # So far I've only encountered this when AD auth legitimately fails or the user is locked/disabled...
            logger.error(str(e))
            raise exceptions.AuthenticationFailed(
                f"Failed to authenticate user {username}. Ensure the credentials are correct & the "
                f"user is not locked/disabled {connection_display}"
            ) from None

        elif "adaptive server is unavailable or does not exist" in error_message:
            # So far I've only encountered this when there are network issues or the
            # instance & port combination is incorrect...
            logger.error(str(e))
            raise exceptions.ConnectionFailed(
                f"Failed to connect to server. This may be a connectivity issue, or an incorrect "
                f"combination of instance & port {connection_display}"
            ) from None
        else:
            logger.error("an unhandled exception was raised")
            raise e
    finally:
        if connection:
            connection.close()


def _execute(connection, *, query: str, values: tuple = ()):
    try:
        cursor = connection.cursor()
        cursor.execute(query, values)
    except (pyodbc.ProgrammingError, pyodbc.Error) as e:

        # Exceptions from pyodbc contain an error code and message. Unfortunately, some error codes are reused so we
        # need to check the message
        error_message = str(e)
        logger.exception(error_message)
        try:
            raise EXECUTE_ERRORS[error_message.lower()]
        except KeyError:
            if "The driver did not supply an error!" in error_message and "?" in query and not values:
                raise exceptions.InvalidSQLQuery("No values were provided for the placeholders in the SQL Query.")
            else:
                raise exceptions.InvalidSQLQuery("Executing SQL Query failed. Please check the logs.")
    return cursor


def _check_for_sql_injection(query, exception_message):
    delimiter_count = query.count(";")
    if delimiter_count == 1 and not query.endswith(";"):
        raise exceptions.SQLInjection(exception_message)
    elif delimiter_count > 1:
        raise exceptions.SQLInjection(exception_message)


def _remove_comments(query):
    # Remove comments from SQL query for our SQLInjection checks
    query = re.sub(r"/\*[^*]*(?:\*(?!/)[^*]*)*\*/", "", query, flags=re.MULTILINE)  # Remove multi line comments
    # Remove -- comments
    lines = query.splitlines()
    return " ".join([line.split("--")[0].strip() for line in lines])
