import functools
from typing import Optional, Union
from typeguard import typechecked
from . import _client


@typechecked
def initialize(
    *,
    address: str,
    default_database: str,
    username: str,
    password: str,
    instance: Optional[str] = None,
    port: Optional[Union[str, int]] = None,
    freetds_driver: Optional[str] = "/usr/lib/libtdsodbc.so.0",
):
    """ Initialize an MSSQL session and get a connector context manager.

        :param address: The hostname or IP address of the MSSQL server to connect to.
        :param instance: The MSSQL server instance name.
        :param default_database: The default database to connect to.
        :param username: The username to log in with.
        :param password: The password to log in with.
        :param port: Optional port.
        :param freetds_driver: Optional. Defaults to the freetds driver location on Alpine Linux. If you are using a
            different base image, you may need to specify a different location.
        :returns: A context manager that yields an opened MSSQL connection.

        The returned context manager should only be called using a `with statement
        <https://docs.python.org/3.8/reference/compound_stmts.html#with>`_.

        All operations must happen inside the `with` block. You do not need to know how to
        open or close an MSSQL connection - the context manager handles all of this for you.

        You can re-use an initialized session as often as necessary. This makes it easier to separate
        offline programming logic from MSSQL module operations as much as possible.

        See the example below to view it in action.

        .. code-block:: python

            sql_adaptor = client.initialize(
                address="10.10.2.1",
                instance="SQLEXPRESS",
                database_name="master"
                username="neil.sql",
                password="jhgf28f0sjicjosdf",
            )

            with sql_adaptor() as connection:
                # Now you can perform an operation against the Microsoft SQL Server
                client.read(connection, query="SELECT * FROM users;")
    """
    return functools.partial(
        _client._connect,
        address=address,
        default_database=default_database,
        username=username,
        password=password,
        instance=instance,
        port=port,
        freetds_driver=freetds_driver,
    )
