from typing import TypedDict, Union, Optional


class MSSQLServer(TypedDict):
    r"""Contains MSSQL Server connection details.

!!! warning "Supplying Credentials"

    You should ___always___ use a PPA Vault integration to provide credentials to a plugin action.

!!! info "Authentication Methods"

    Both local SQL & Active Directory authentication are supported.

    For Active Directory authentication the username must be in this format:

    `domain\username`

    The domain portion should be the short-format name (NETBIOS Domain Name).

!!! danger "Required Keys"

    __address__: Server IP or DNS address

    __database__: Database to connect to

    __username__: Username for authentication

    __password__: Password for authentication

!!! info "Optional Keys"

    __instance__: The SQL database instance to connect to.

    __port__: The port to use when connecting to the instance.

    Depending on what PPA needs to connect to, at least one of these values will be required.

    See the table below for more information.

    The plugin will fail if neither is supplied.

When To Supply Instance & Port

| Multiple Instance Server | Default Instance | Default Database | Required Details                          |
| ------------------------ | ---------------- | ---------------- | ----------------------------------------- |
| No                       | Yes              | Yes              | **Instance** _and/or_ **Port**            |
| No                       | Yes              | No               | **Port** _or_ **Port** & **Instance**     |
| Yes                      | Yes              | Yes              | **Instance** _and/or_ **Port**            |
| Yes                      | Yes              | No               | **Instance** _and/or_ **Port**            |
| Yes                      | No               | Yes              | **Instance** _and/or_ **Port**            |
| Yes                      | No               | No               | **Instance** _only_                       |

??? example "Examples - Single-Instance Server"

    Connecting to the **default database**:

    ```yaml linenums="1" hl_lines="6 7 8"
    mssql:
      address: sql-server.internal.net
      database: master
      username: example.username
      password: example.password
      # Either or both of the following is accepted
      instance: SQLEXPRESS
      port: 1433
    ```

    Connecting to a **non-default_ database**:

    ```yaml linenums="1" hl_lines="6 7 8"
    mssql:
      address: sql-server.internal.net
      database: master
      username: example.username
      password: example.password
      # The port must be supplied, the instance name is optional
      port: 1433
      instance: SQLEXPRESS
    ```

??? example "Examples - Multiple-Instance Server"

    Connecting to **any database** on the **default instance**:

    ```yaml linenums="1" hl_lines="6 7 8"
    mssql:
      address: sql-server.internal.net
      database: master
      username: example.username
      password: example.password
      # Either or both of the following is accepted
      instance: SQLEXPRESS
      port: 1433
    ```

    Connecting to the **default database** on a **non-default instance**:

    ```yaml linenums="1" hl_lines="6 7 8"
    mssql:
      address: sql-server.internal.net
      database: master
      username: example.username
      password: example.password
      # Either or both of the following is accepted
      instance: PAYROLL
      port: 1433
    ```

    Connecting to a **non-default database** on a **non-default instance**:

    ```yaml linenums="1" hl_lines="6 7"
    mssql:
      address: sql-server.internal.net
      database: users
      username: example.username
      password: example.password
      # Only the instance must be supplied
      instance: PAYROLL
    ```

    """
    address: str
    database: str
    username: str
    password: str
    instance: Optional[str] = None
    port: Optional[Union[int, str]] = None
